@@ -4,7 +4,7 @@
  * See http://mod-qos.sourceforge.net/ for further
  * details.
  *
- * Copyright (C) 2022 Pascal Buchbinder
+ * Copyright (C) 2023 Pascal Buchbinder
  *
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
@@ -29,8 +29,8 @@
 /* ----------------------------------
  * version info
  * ---------------------------------- */
-static const char man_version[] = "11.72";
-static const char man_date[] = "May 2022";
+static const char man_version[] = "11.73";
+static const char man_date[] = "January 2023";
 
 /* ----------------------------------
  * definitions
@@ -71,4 +71,28 @@ void qs_deleteOldFiles(const char *file_name, int generations);
 /* user */
 void qs_setuid(const char *username, const char *cmd);
 
+/* pcre */
+#define QS_MAX_REG_MATCH 10
+
+typedef struct {
+  int rm_so;
+  int rm_eo;
+} qs_regmatch_t;
+
+typedef struct {
+  void *re_pcre;
+  int re_nsub;
+  int state; 
+} qs_regex_t;
+
+int qs_pregfree(void *p);
+
+void qs_regfree(qs_regex_t *preg);
+
+int qs_regcomp(qs_regex_t *preg, const char *regex, int cflags);
+
+int qs_regexec_len(const qs_regex_t *preg, const char *buff,
+		   unsigned int len, unsigned int nmatch,
+		   qs_regmatch_t *pmatch, int eflags);
+
 #endif
