@@ -5,7 +5,7 @@
  *
  * See http://mod-qos.sourceforge.net/ for further details.
  *
- * Copyright (C) 2022 Pascal Buchbinder
+ * Copyright (C) 2023 Pascal Buchbinder
  *
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
@@ -35,7 +35,6 @@ static const char revision[] = "$Id: qsexec.c 2654 2022-05-13 09:12:42Z pbuchbin
 #include <time.h>
 
 /* apr */
-#include <pcre.h>
 #include <apr.h>
 #include <apr_strings.h>
 #include <apr_file_io.h>
@@ -46,12 +45,14 @@ static const char revision[] = "$Id: qsexec.c 2654 2022-05-13 09:12:42Z pbuchbin
 #include <apr_portable.h>
 #include <apr_support.h>
 
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
+
 #include "qs_util.h"
 
 #ifndef POSIX_MALLOC_THRESHOLD
 #define POSIX_MALLOC_THRESHOLD (10)
 #endif
-#define MAX_REG_MATCH 10
 
 /* same as APR_SIZE_MAX which doesn't appear until APR 1.3 */
 #define QSUTIL_SIZE_MAX (~((apr_size_t)0))
@@ -154,7 +155,7 @@ static void usage(char *cmd, int man) {
  */
 char *qs_pregsub(apr_pool_t *pool, const char *input,
 		 const char *source, size_t nmatch,
-		 regmatch_t pmatch[]) {
+		 qs_regmatch_t pmatch[]) {
   const char *src = input;
   char *dest, *dst;
   char c;
@@ -216,41 +217,6 @@ char *qs_pregsub(apr_pool_t *pool, const char *input,
   return dest;
 }
 
-int qs_regexec(pcre *preg, const char *string,
-	       apr_size_t nmatch, regmatch_t pmatch[]) {
-  int rc;
-  int options = 0;
-  int *ovector = NULL;
-  int small_ovector[POSIX_MALLOC_THRESHOLD * 3];
-  int allocated_ovector = 0;
-  if (nmatch > 0) {
-    if (nmatch <= POSIX_MALLOC_THRESHOLD) {
-      ovector = &(small_ovector[0]);
-    } else {
-      ovector = (int *)malloc(sizeof(int) * nmatch * 3);
-      if (ovector == NULL) {
-	return 1;
-      }
-      allocated_ovector = 1;
-    }
-  }
-  rc = pcre_exec(preg, NULL, string, (int)strlen(string), 0, options, ovector, nmatch * 3);
-  if (rc == 0) rc = nmatch;    /* All captured slots were filled in */
-  if (rc >= 0) {
-    apr_size_t i;
-    for (i = 0; i < (apr_size_t)rc; i++) {
-      pmatch[i].rm_so = ovector[i*2];
-      pmatch[i].rm_eo = ovector[i*2+1];
-    }
-    if (allocated_ovector) free(ovector);
-    for (; i < nmatch; i++) pmatch[i].rm_so = pmatch[i].rm_eo = -1;
-    return 0;
-  } else {
-    if (allocated_ovector) free(ovector);
-    return rc;
-  }
-}
-
 int main(int argc, const char * const argv[]) {
   const char *username = NULL;
   int nr = 0;
@@ -262,11 +228,9 @@ int main(int argc, const char * const argv[]) {
   const char *clearcommand = NULL;
   const char *clearpattern = NULL;
   int executed = 0;
-  pcre *preg;
-  pcre *clearpreg;
-  const char *errptr = NULL;
-  int erroffset;
-  regmatch_t regm[MAX_REG_MATCH];
+  qs_regex_t *preg;
+  qs_regex_t *clearpreg;
+  qs_regmatch_t regm[QS_MAX_REG_MATCH];
   time_t sec = 0;
   int threshold = 0;
   int counter = 0;
@@ -342,32 +306,36 @@ int main(int argc, const char * const argv[]) {
 
   qs_setuid(username, cmd);
 
-  preg = pcre_compile(pattern, PCRE_DOTALL, &errptr, &erroffset, NULL);
-  if(!preg) {
-    fprintf(stderr, "ERROR, could not compile '%s' at position %d, reason: %s\n",
-	    pattern, erroffset, errptr);
+  preg = apr_palloc(pool, sizeof *preg);
+  if(qs_regcomp(preg, pattern, PCRE2_DOTALL) != 0) {
+    fprintf(stderr, "ERROR, could not compile '%s'\n", pattern);
     exit(1);
   }
+  apr_pool_pre_cleanup_register(pool, preg, qs_pregfree);
   if(clearpattern) {
-    clearpreg = pcre_compile(clearpattern, PCRE_DOTALL, &errptr, &erroffset, NULL);
-    if(!clearpreg) {
-      fprintf(stderr, "ERROR, could not compile '%s' at position %d, reason: %s\n",
-	      clearpattern, erroffset, errptr);
+    clearpreg = apr_palloc(pool, sizeof *clearpreg);
+    if(qs_regcomp(clearpreg, clearpattern, PCRE2_DOTALL) != 0) {
+      fprintf(stderr, "ERROR, could not compile '%s'\n", clearpattern);
       exit(1);
     }
+    apr_pool_pre_cleanup_register(pool, clearpattern, qs_pregfree);
   }
 
   while(fgets(line, MAX_LINE_BUFFER, stdin) != NULL) {
+    size_t len;
     nr++;
     if(pass) {
       printf("%s", line);
       fflush(stdout);
     }
-    if(clearpattern && (qs_regexec(clearpreg, line, MAX_REG_MATCH, regm) == 0)) {
+    len = strlen(line);
+    if(clearpattern && (qs_regexec_len(clearpreg, line, len, QS_MAX_REG_MATCH, regm, 0) >= 0)) {
+      apr_pool_t *subpool;
+      apr_pool_create(&subpool, pool);
       counter = 0;
       countertime = 0;
       if(clearcommand && executed) {
-	char *replaced = qs_pregsub(pool, clearcommand, line, MAX_REG_MATCH, regm);
+	char *replaced = qs_pregsub(subpool, clearcommand, line, QS_MAX_REG_MATCH, regm);
 	if(!replaced) {
 	  fprintf(stderr, "[%s]: ERROR, failed to substitute"
                   " submatches '%s' in (%s)\n", cmd, clearcommand, line);
@@ -376,8 +344,12 @@ int main(int argc, const char * const argv[]) {
 	}
 	executed = 0;
       }
-    } else if(qs_regexec(preg, line, MAX_REG_MATCH, regm) == 0) {
-      char *replaced = qs_pregsub(pool, command, line, MAX_REG_MATCH, regm);
+      apr_pool_destroy(subpool);
+    } else if(qs_regexec_len(preg, line, len, QS_MAX_REG_MATCH, regm, 0) >= 0) {
+      apr_pool_t *subpool;
+      char *replaced;
+      apr_pool_create(&subpool, pool);
+      replaced = qs_pregsub(subpool, command, line, QS_MAX_REG_MATCH, regm);
       if(!replaced) {
 	fprintf(stderr, "[%s]: ERROR, failed to substitute"
                 " submatches '%s' in (%s)\n", cmd, command, line);
@@ -395,7 +367,7 @@ int main(int argc, const char * const argv[]) {
 	  counter = 0;
 	}
       }
-      apr_pool_clear(pool);
+      apr_pool_destroy(subpool);
     }
   }
   
