@@ -7,7 +7,7 @@
  *
  * See http://mod-qos.sourceforge.net/ for further details.
  *
- * Copyright (C) 2022 Pascal Buchbinder
+ * Copyright (C) 2023 Pascal Buchbinder
  *
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
@@ -38,7 +38,6 @@ static const char revision[] = "$Id: qsgrep.c 2654 2022-05-13 09:12:42Z pbuchbin
 #include <time.h>
 
 /* apr */
-#include <pcre.h>
 #include <apr.h>
 #include <apr_strings.h>
 #include <apr_file_io.h>
@@ -49,12 +48,15 @@ static const char revision[] = "$Id: qsgrep.c 2654 2022-05-13 09:12:42Z pbuchbin
 #include <apr_portable.h>
 #include <apr_support.h>
 
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
+
 #include "qs_util.h"
 
 #ifndef POSIX_MALLOC_THRESHOLD
 #define POSIX_MALLOC_THRESHOLD (10)
 #endif
-#define MAX_REG_MATCH 10
+
 /* same as APR_SIZE_MAX which doesn't appear until APR 1.3 */
 #define QSUTIL_SIZE_MAX (~((apr_size_t)0))
 
@@ -140,7 +142,7 @@ static void usage(char *cmd, int man) {
  */
 char *qs_pregsub(apr_pool_t *pool, const char *input,
 		 const char *source, size_t nmatch,
-		 regmatch_t pmatch[]) {
+		 qs_regmatch_t pmatch[]) {
   const char *src = input;
   char *dest, *dst;
   char c;
@@ -202,41 +204,6 @@ char *qs_pregsub(apr_pool_t *pool, const char *input,
   return dest;
 }
 
-int qs_regexec(pcre *preg, const char *string,
-	       apr_size_t nmatch, regmatch_t pmatch[]) {
-  int rc;
-  int options = 0;
-  int *ovector = NULL;
-  int small_ovector[POSIX_MALLOC_THRESHOLD * 3];
-  int allocated_ovector = 0;
-  if (nmatch > 0) {
-    if (nmatch <= POSIX_MALLOC_THRESHOLD) {
-      ovector = &(small_ovector[0]);
-    } else {
-      ovector = (int *)malloc(sizeof(int) * nmatch * 3);
-      if (ovector == NULL) {
-	return 1;
-      }
-      allocated_ovector = 1;
-    }
-  }
-  rc = pcre_exec(preg, NULL, string, (int)strlen(string), 0, options, ovector, nmatch * 3);
-  if (rc == 0) rc = nmatch;    /* All captured slots were filled in */
-  if (rc >= 0) {
-    apr_size_t i;
-    for (i = 0; i < (apr_size_t)rc; i++) {
-      pmatch[i].rm_so = ovector[i*2];
-      pmatch[i].rm_eo = ovector[i*2+1];
-    }
-    if (allocated_ovector) free(ovector);
-    for (; i < nmatch; i++) pmatch[i].rm_so = pmatch[i].rm_eo = -1;
-    return 0;
-  } else {
-    if (allocated_ovector) free(ovector);
-    return rc;
-  }
-}
-
 int main(int argc, const char * const argv[]) {
   unsigned long nr = 0;
   char line[32768];
@@ -246,10 +213,8 @@ int main(int argc, const char * const argv[]) {
   const char *out = NULL;
   const char *pattern = NULL;
   const char *filename = NULL;
-  pcre *preg;
-  const char *errptr = NULL;
-  int erroffset;
-  regmatch_t regm[MAX_REG_MATCH];
+  qs_regex_t *preg;
+  qs_regmatch_t regm[QS_MAX_REG_MATCH];
   apr_app_initialize(&argc, &argv, NULL);
   apr_pool_create(&pool, NULL);
 
@@ -293,12 +258,12 @@ int main(int argc, const char * const argv[]) {
     fprintf(stderr, "ERROR, failed to change nice value: %s\n", strerror(errno));
   }
 
-  preg = pcre_compile(pattern, PCRE_DOTALL, &errptr, &erroffset, NULL);
-  if(!preg) {
-    fprintf(stderr, "ERROR, could not compile '%s' at position %d, reason: %s\n",
-	    pattern, erroffset, errptr);
+  preg = apr_palloc(pool, sizeof *preg);
+  if(qs_regcomp(preg, pattern, PCRE2_DOTALL) != 0) {
+    fprintf(stderr, "ERROR, could not compile '%s\n", pattern);
     exit(1);
   }
+  apr_pool_pre_cleanup_register(pool, preg, qs_pregfree);
 
   if(filename) {
     file = fopen(filename, "r");
@@ -311,16 +276,20 @@ int main(int argc, const char * const argv[]) {
   }
     
   while(fgets(line, sizeof(line), file) != NULL) {
+    size_t len = strlen(line);
     nr++;
-    if(qs_regexec(preg, line, MAX_REG_MATCH, regm) == 0) {
-      char *replaced = qs_pregsub(pool, out, line, MAX_REG_MATCH, regm);
+    if(qs_regexec_len(preg, line, len, QS_MAX_REG_MATCH, regm, 0) >= 0) {
+      apr_pool_t *subpool;
+      char *replaced;
+      apr_pool_create(&subpool, pool);
+      replaced = qs_pregsub(subpool, out, line, QS_MAX_REG_MATCH, regm);
       if(!replaced) {
 	fprintf(stderr, "ERROR, failed to substitute submatches (line=%lu)\n", nr);
       } else {
 	printf("%s\n", replaced);
         fflush(stdout);
       }
-      apr_pool_clear(pool);
+      apr_pool_destroy(subpool);
     }
   }
 
