@@ -9,7 +9,7 @@
  * See http://mod-qos.sourceforge.net/ for further
  * details.
  *
- * Copyright (C) 2022 Pascal Buchbinder
+ * Copyright (C) 2023 Pascal Buchbinder
  *
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
@@ -45,14 +45,15 @@ static const char revision[] = "$Id: qslog.c 2654 2022-05-13 09:12:42Z pbuchbind
 #include <regex.h>
 #include <time.h>
 
-#include <pcre.h>
-
 /* apr */
 #include <apr.h>
 #include <apr_portable.h>
 #include <apr_support.h>
 #include <apr_strings.h>
 
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
+
 #include "qs_util.h"
 
 /* ----------------------------------
@@ -990,31 +991,28 @@ static void qs_updateEvents(apr_pool_t *pool, char *E, apr_table_t *events) {
 static void qsInitCounter(apr_pool_t *pool, apr_table_t *counters) {
   const char *envFile = getenv(QSCOUNTERPATH);
   if(envFile != NULL) {
-    const char *errptr = NULL;
-    int erroffset;
-    int ovector[100];
-    pcre *pcrestat = pcre_compile(COUNTER_PATTERN, PCRE_CASELESS, &errptr, &erroffset, NULL);
+    qs_regmatch_t regm[QS_MAX_REG_MATCH];
+    qs_regex_t *pcrestat = apr_palloc(pool, sizeof *pcrestat);
     FILE *file = fopen(envFile, "r"); 
+    if(qs_regcomp(pcrestat, COUNTER_PATTERN, PCRE2_CASELESS) != 0) {
+      fprintf(stderr, "ERROR, could not compile '%s'\n", COUNTER_PATTERN);
+      exit(1);
+    }
+    apr_pool_pre_cleanup_register(pool, pcrestat, qs_pregfree);
     if(file != NULL) {
       char line[MAX_LINE];
       while(!qs_getLinef(line, sizeof(line), file)) {
-        if(pcre_exec(pcrestat, NULL, line, strlen(line), 0, 0, ovector, 100) >= 0) {
+        if(qs_regexec_len(pcrestat, line, strlen(line), QS_MAX_REG_MATCH, regm, 0) >= 0) {
           counter_rec_t *c = apr_pcalloc(pool, sizeof(counter_rec_t));
-          line[ovector[2] + ovector[3] - ovector[2]] = '\0';
-          line[ovector[4] + ovector[5] - ovector[4]] = '\0';
-          line[ovector[6] + ovector[7] - ovector[6]] = '\0';
-          line[ovector[8] + ovector[9] - ovector[8]] = '\0';
-          line[ovector[10] + ovector[11] - ovector[10]] = '\0';
- 
-          c->name = apr_pstrdup(pool, &line[ovector[2]]);
+          c->name = apr_pstrndup(pool, &line[regm[1].rm_so], regm[1].rm_eo - regm[1].rm_so);
           c->count = 0;
           c->total = 0;
           c->start = 0;
-          c->inc = apr_pstrdup(pool, &line[ovector[4]]);
-          c->decVal = atoi(&line[ovector[6]]);
-          c->dec = apr_pstrdup(pool, &line[ovector[8]]);
-          c->duration = atoi(&line[ovector[10]]);
-          c->limit = atoi(&line[ovector[12]]);
+          c->inc = apr_pstrndup(pool, &line[regm[2].rm_so], regm[2].rm_eo - regm[2].rm_so);
+          c->decVal = atoi(apr_pstrndup(pool, &line[regm[3].rm_so], regm[3].rm_eo - regm[3].rm_so));
+          c->dec = apr_pstrndup(pool, &line[regm[4].rm_so], regm[4].rm_eo - regm[4].rm_so);
+          c->duration = atoi(apr_pstrndup(pool, &line[regm[5].rm_so], regm[5].rm_eo - regm[5].rm_so));
+          c->limit = atoi(apr_pstrndup(pool, &line[regm[6].rm_so], regm[6].rm_eo - regm[6].rm_so));
           if(m_verbose) {
             fprintf(stderr, "%s : %s - (%d * %s) / %d = %d\n",
                     c->name, c->inc, c->decVal, c->dec, c->duration, c->limit);
